#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from importlib import import_module
from os.path import splitext

from pandas import DataFrame


MODULE_CACHE = {}


def smart_reader(file_path, *, vanilla=False):
    """Dispatch a file reader based on file extension.

    Parameters
    ----------
    file_path : str
        Path to the file to be read.
    vanilla : bool, optional
        Return a spruced up subclass of the `pandas.DataFrame` (`ActivityData`)
        and benefit from some extra data pruning and functionality, or be
        boring and get the raw data untouched.

    Returns
    -------
    ActivityData or DataFrame
        The output depends on the `vanilla` argument.

    Raises
    ------
    ImportError
        If the file type (based on the extension) is not supported.
    """
    ext = splitext(file_path)[-1][1:]   # drop period from the extension

    module = MODULE_CACHE.get(ext, None)
    if module is None:
        try:
            MODULE_CACHE[ext] = import_module('activityio.' + ext)
        except ImportError:
            raise ImportError('%s is not a supported file type' % ext)
        else:
            module = MODULE_CACHE.get(ext)

    if not vanilla:
        return module.read(file_path)
    else:
        return DataFrame.from_records(module.gen_records(file_path))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Some useful functions for parsing XML file types.

Note we need to name this `xml_reading` so as not to clobber the standard
library package.

"""
from xml.etree.cElementTree import iterparse

def gen_nodes(file_path, node_names, *, with_root=False):
    """Efficiently iterate over specific nodes of an XML document.

    http://effbot.org/zone/element-iterparse.htm
    """
    context = iter(iterparse(file_path, events=('start', 'end')))
    event, root = next(context)  # get the root element

    if with_root:
        yield root

    for event, element in context:
        if event == 'end' and sans_ns(element.tag) in node_names:
            yield element
            root.clear()


def sans_ns(tag):
    """Remove the namespace prefix from a tag."""
    return tag.split('}')[-1]



"""
Decode Schoberer Rad Messtechnik (SRM) power files.

This code was adapted largely from the GoldenCheetah code base [1]_.

Rainer Clasen [2]_, developer of the `srmio` utility, was also kind enough to
provide a file format specification. This specification, though incomplete, can
be found in the '_references' directory of this subpackage.


.. [1] https://github.com/GoldenCheetah/GoldenCheetah/blob/master/src/FileIO/SrmRideFile.cpp
.. [2] https://github.com/rclasen

"""
from activityio.srm._reading import read_and_format as read
from activityio.srm._reading import gen_records



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Exceptions for this package.

"""
class ActivityIOError(Exception):
    """The base exception for all this package's exceptions."""
    _default_message = ''

    def __init__(self, message=None):
        super().__init__(message if message else self._default_message)


class InvalidFileError(Exception):
    _default_message = 'unexpected file format'




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A bonus module for prettifying console output.

"""
import sys


TEXT_DECORATIONS = {
    'header': '\033[95m',
    'blue': '\033[94m',
    'green': '\033[92m',
    'warning': '\033[93m',
    'fail': '\033[91m',
    'bold': '\033[1m',
    'underline': '\033[4m',
    'end': '\033[0m',
}


class indented_stdout:
    """Context manager for indenting anything sent to stdout.

        >>> with indented_stdout(2):
        ...    print('this is indented')
        ...
          this is intended
    """
    def __init__(self, indent=4):
        self.indent = ' ' * indent
        self.should_indent = True

    def write(self, text):
        if self.should_indent:
            text = self.indent + text
        # https://docs.python.org/3/library/sys.html#sys.__stdout__
        sys.__stdout__.write(text)
        self.should_indent = text.endswith('\n')    # for next time

    def __enter__(self):
        sys.stdout = self
        return self

    def __exit__(self, type, value, traceback):
        sys.stdout = sys.__stdout__


def decorate(text, *decorations):
    """Return a text string with ANSI escape codes pre- and appended.

    Parameters
    ----------
    text : str
        Text to be decorated.
    *decorations : str
        Currently supported escape codes are:
            + header
            + blue
            + green
            + warning
            + fail
            + bold
            + underline
            + end
    """
    decors = ''.join(TEXT_DECORATIONS[d] for d in decorations)
    end = TEXT_DECORATIONS['end']
    return decors + text + end


def printd(text, *decorations, **kwargs):
    """Print decorated."""
    to_write = decorate(text, *decorations)
    print(to_write, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
General utilities to be used internally.

"""
import numpy as np


def transformer(functions):
    """Apply functions to a pandas.DataFrame by column name

    Returned function should be passed to the ``apply`` method of DataFrames.

    http://stackoverflow.com/questions/26434123

    Parameters
    ----------
    functions : dict
        A dictionary of functions, the keys for which determine the columns
        to which those functions should be applied.
    """
    def func(col):
        if col.name not in functions:
            return col
        else:
            return functions[col.name](col)
    return func


def zip_staggered(iterable):
    ls = list(iterable)   # need to consume the iterable, sadly
    return zip(ls[:-1], ls[1:])


def make_lap_column(length, new_lap_i):
    laps = np.zeros(length)
    laps[new_lap_i] = 1
    return laps.cumsum().astype(np.int64)


def semicircles_to_degrees(semicircles):
    """Positional data conversion for *.fit files

    https://github.com/kuperov/fit/blob/master/R/fit.R
    """
    return (semicircles * 180 / 2**31 + 180) % 360 - 180



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""


"""
from datetime import datetime, timedelta
from itertools import islice

from activityio._util.xml_reading import gen_nodes, sans_ns
from activityio._util import drydoc, types
from activityio._util.exceptions import InvalidFileError


DATETIME_FMT = '%Y-%m-%dT%H:%M:%S'    # timezone unspecified


COLUMN_SPEC = {
    'alt': types.Altitude,
    'cad': types.Cadence,
    'dist': types.Distance,
    'hr': types.HeartRate,
    'pwr': types.Power,
    'spd': types.Speed,
    'temp': types.Temperature,
}


def format_sample(sample):
    return {sans_ns(child.tag): float(child.text) for child in
            # Need to ignore the first (sample) element
            islice(sample.iter(), 1, None)}


@drydoc.gen_records
def gen_records(file_path):
    find_these = ['time', 'sample']
    nodes = gen_nodes(file_path, find_these, with_root=True)

    root = next(nodes)
    if sans_ns(root.tag) != 'pwx':
        raise InvalidFileError("this doesn't look like a pwx file!")

    start_time = datetime.strptime(next(nodes).text, DATETIME_FMT)
    find_these.pop(0)

    samples = nodes

    for sample in samples:
        sample_dict = format_sample(sample)
        sample_dict['timestamp'] = (
            start_time + timedelta(seconds=sample_dict['timeoffset']))
        yield sample_dict


def read_and_format(file_path):
    data = types.ActivityData.from_records(gen_records(file_path))

    timestamps = data.pop('timestamp')
    timeoffsets = data.pop('timeoffset')
    data._finish_up(column_spec=COLUMN_SPEC,
                    start=timestamps[0], timeoffsets=timeoffsets)

    return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Avoid repeating documentation.

These functions are intended as decorators, which will pass their own docstring
to the incoming function. Neat!

"""
import inspect

def gen_records(func):
    """Generator function for iterating over individual file records.

    "Records" are dictionary objects representing a single "sample" of data;
    i.e. a row in a tabular representation. Note this can be passed to
    the `from_records` constructor method of `pandas.DataFrame`s.
    """
    this_func = inspect.stack()[0][3]
    this_doc = globals().get(this_func).__doc__
    func.__doc__ = this_doc
    return func



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from argparse import ArgumentParser
from functools import partial
from importlib import import_module
from os import path, listdir

import activityio


VALID_FORMATS = tuple(d for d in listdir(path.dirname(activityio.__file__))
                      if not d.startswith('_') and not d.endswith('py'))


def main():

    # Argument handling
    # -----------------
    parser = ArgumentParser(description='parse an activity file')

    parser.add_argument('input',
                        type=str,
                        help='raw file to read')
    parser.add_argument('--output',
                        type=str,
                        metavar='filename',
                        default=None,
                        help='optional; file to write to')
    parser.add_argument('--format',
                        type=str,
                        default=None,
                        help='optional; format of the file',
                        choices=VALID_FORMATS)

    file_path, output, fmt = parser.parse_args()

    # Script begins
    # -------------
    if fmt is None:
        fmt = path.splitext(file_path)[-1][1:]

    module = import_module('activityio.' + fmt)
    data = module.read(file_path)
    write = partial(data.to_csv,
                    na_rep='NA', index_label='time', encoding='utf-8')
    if output is None:
        csv = write()
        print(csv)
    else:
        write(output)

    return 0


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from functools import wraps

import numpy as np
from pandas import DataFrame, Series, Timedelta, TimedeltaIndex

from activityio import tools
from activityio._util.exceptions import ActivityIOError


SPECIAL_COLUMNS = {}    # grows on import via @special decorator


class RequiredColumnError(ActivityIOError):
    def __init__(self, column, cls=None):
        if cls is None:
            message = '{!r} column not found'.format(column)
        else:
            message = '{!r} column should be of type {!s}'.format(column, cls)
        super().__init__(message)


def new_column_sugar(needs: tuple, name=None):
    """Decorator for certain methods of ActivityData that create new columns
    using the special column types specified here.

    Parameters
    ----------
    needs : tuple
        A tuple of colnames.
    name : str, optional
        The name for the returned Series object.

    Returns
    -------
    Series
        Suitable for joining to the existing data.

    Raises
    ------
    RequiredColumnError
        If a column specified in `needs` is not present.
    """
    def real_decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            for need in needs:
                if need not in self:
                    raise RequiredColumnError(need)

            # Fine, so construct the new Series.
            out = func(self, *args, **kwargs)
            return Series(out, index=self.index, name=name)
        return wrapper
    return real_decorator


class ActivityData(DataFrame):
    _metadata = ['start', '_athlete']

    @property
    def _constructor(self):
        return ActivityData

    def __finalize__(self, other, method=None, **kwargs):
        """Propagate metadata from other to self."""
        for name in self._metadata:
            object.__setattr__(self, name, getattr(other, name, None))
        return self

    def __getitem__(self, key):
        """Create the illusion of Series subclasses in the DataFrame."""
        item = super().__getitem__(key)
        try:
            return SPECIAL_COLUMNS[key](item)
        except KeyError:
            return item

    def _finish_up(self, *, column_spec, start=None, timeoffsets=None):
        for old_key, column_cls in column_spec.items():
            # DataFrame.pop() does not take a default arg
            try:
                old_column = self.pop(old_key)
            except KeyError:
                continue

            new = column_cls(old_column)
            self[new.colname] = new

        self.start = start
        if timeoffsets is not None:
            self.index = TimedeltaIndex(timeoffsets, unit='s', name='time')

        # No point hanging on to completely empty columns!
        self.dropna(axis=1, how='all', inplace=True)

    @property
    def time(self):   # makes accessing the index more readable
        if isinstance(self.index, TimedeltaIndex):
            return self.index
        else:
            # because recursion problems with super().__getattr__()
            raise AttributeError('index is not TimedeltaIndex')

    def rollmean(self, column, seconds):
        """Apply rolling mean by time to column."""
        return self.resample_1hz()[column].rolling(seconds).mean()

    def resample_1hz(self):
        """Resample to 1 Hz, filling gaps with NaNs."""
        return self.resample('1s').mean()

    def normpwr(self):
        """Training Peaks 'Normalised Power' (NP) metric."""
        try:
            smooth_pwr = self.rollmean('pwr', 30)
        except KeyError as e:
            raise RequiredColumnError('pwr') from e

        return np.mean(smooth_pwr**4)**(1/4)

    @new_column_sugar(needs=('lon', 'lat'), name='dists_m')
    def haversine(self, **kwargs):
        lon, lat = (self[ax].radians.values for ax in ('lon', 'lat'))
        return tools.haversine(lon, lat, **kwargs)

    @new_column_sugar(needs=('lon', 'lat'), name='bearing_deg')
    def bearing(self, **kwargs):
        lon, lat = (self[ax].radians.values for ax in ('lon', 'lat'))
        return tools.bearing(lon, lat, **kwargs)

    def vam(self):
        dtime = np.diff(self.index.total_seconds())
        # numpy.diff doesn't prepend an NaN
        dtime = np.insert(dtime, 0, np.nan)
        dvert = self['alt'].diff()
        return VAM(dvert / dtime)

    def gradient(self):
        needs = ('alt', 'dist')
        for need in needs:
            if need not in self:
                raise RequiredColumnError(need)

        return Gradient(rise=self['alt'].diff().values,
                        run=self['dist'].diff().values)


class SeriesSubclass(Series):
    _metadata = ['colname', 'base_unit']

    @property
    def _constructor(self):
        return self.__class__

    def __finalize__(self, other, method=None, **kwargs):
        """Propagate metadata from other to self."""
        for name in self._metadata:
            object.__setattr__(self, name, getattr(other, name, None))
        return self

    def __init__(self, data, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        self._name = self.__class__.colname     # use class attribute

    # NOTE: subclasses should follow the structure...
    #   + classmethods (i.e. alternative constructors; private!)
    #   + general methods
    #   + properties


def special(cls):
    """Decorator for easily growing the SPECIAL_COLUMNS global dict."""
    global SPECIAL_COLUMNS
    SPECIAL_COLUMNS[cls.colname] = cls
    return cls


class series_property:
    """A simple descriptor that emulates property, but returns a Series."""
    def __init__(self, fget):
        self.fget = fget
        self.__doc__ = fget.__doc__

    def __get__(self, obj, objtype=None):
        return Series(self.fget(obj))


@special
class Altitude(SeriesSubclass):
    colname = 'alt'
    base_unit = 'm'

    @property
    def ascent(self):
        deltas = self.diff()
        cls = type(self)
        return cls(np.where(deltas > 0, deltas, 0))

    @property
    def descent(self):
        deltas = self.diff()
        cls = type(self)
        return cls(np.where(deltas < 0, deltas, 0))

    @series_property
    def ft(self):
        """ metres --> feet """
        return self * 3.28084


@special
class Cadence(SeriesSubclass):
    colname = 'cad'
    base_unit = 'rpm'


@special
class Distance(SeriesSubclass):
    colname = 'dist'
    base_unit = 'm'

    @classmethod
    def _from_discrete(cls, data, *args, **kwargs):
        return cls(data.cumsum(), *args, **kwargs)

    @series_property
    def km(self):
        """ metres --> kilometres """
        return self / 1000

    @series_property
    def miles(self):
        """ metres --> miles """
        return self / 1000 * 0.621371


@special
class Gradient(SeriesSubclass):
    colname = 'grad'
    base_unit = 'fraction'

    _metadata = ['_rise', '_run'] + SeriesSubclass._metadata

    def __init__(self, *args, rise=None, run=None, **kwargs):
        if rise is not None and run is not None:
            self._rise, self._run = rise, run
            super().__init__(rise/run, *args, **kwargs)
        else:
            super().__init__(*args, **kwargs)

    @series_property
    def pct(self):
        """ fraction --> % """
        return self * 100

    @series_property
    def radians(self):
        """ fraction --> radians """
        return np.arctan2(self._rise, self._run)

    @series_property
    def degrees(self):
        """ fraction --> degrees """
        return np.rad2deg(np.arctan2(self._rise, self._run))


@special
class HeartRate(SeriesSubclass):
    colname = 'hr'
    base_unit = 'bpm'


@special
class LapCounter(SeriesSubclass):
    colname = 'lap'
    base_unit = '#'


class LonLat(SeriesSubclass):
    base_unit = 'degrees'

    @classmethod
    def _from_semicircles(cls, data, *args, **kwargs):
        # https://github.com/kuperov/fit/blob/master/R/fit.R
        deg = (data * 180 / 2**31 + 180) % 360 - 180
        return cls(deg, *args, **kwargs)

    @series_property
    def radians(self):
        """ degrees --> radians """
        return np.radians(self)


@special
class Longitude(LonLat):
    colname = 'lon'


@special
class Latitude(LonLat):
    colname = 'lat'


@special
class Pace(SeriesSubclass):
    colname = 'pace'
    base_unit = 'sec/m'

    @series_property
    def min_per_km(self):
        return self * 1000

    @series_property
    def min_per_mile(self):
        return self * 1000 / 1.61


@special
class Power(SeriesSubclass):
    colname = 'pwr'
    base_unit = 'watts'

    def to_work(self):
        dt = np.diff(self.index.total_seconds())
        dt = np.concatenate(([np.nan], dt))
        return Work(self * dt)

    def wbalance(self, CP, **kwargs):
        wbal = tools.wbalance(power=self.values,
                              timer_sec=self.index.total_seconds(),
                              CP=CP, **kwargs)
        return Series(wbal, index=self.index, name='wbalance')


@special
class Speed(SeriesSubclass):
    colname = 'speed'
    base_unit = 'm/s'

    @classmethod
    def _from_kph(cls, data, *args, **kwargs):
        return cls(data / 60**2 * 1000, *args, **kwargs)

    def to_pace(self):
        tds = (1 / self).apply(Timedelta, args=('s',))
        return Pace(tds)

    @series_property
    def kph(self):
        """ metres/second --> kilometres/hour """
        return self * 60**2 / 1000

    @property
    def mph(self):
        """ metres/second --> miles/hour """
        return self.kph / 1.61    # self.kph is already a Series


@special
class Temperature(SeriesSubclass):
    colname = 'temp'
    base_unit = 'degrees_C'


@special
class VAM(SeriesSubclass):
    colname = 'vam'
    base_unit = 'm/s'


@special
class Work(SeriesSubclass):
    colname = 'work'
    base_unit = 'J'

    @series_property
    def kj(self):
        """ joules --> kilojoules """
        return self / 1000



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Reading logic for SRM files. Should cover versions 5--9.

"""
from collections import deque
from contextlib import contextmanager
from datetime import datetime, timedelta
from functools import reduce
from itertools import accumulate
from math import nan
from operator import ior
from struct import unpack, calcsize

from activityio._util import drydoc, types
from activityio._util.exceptions import InvalidFileError


DATETIME_1880 = datetime(year=1880, month=1, day=1)


COLUMN_SPEC = {
    'alt': types.Altitude,
    'cad': types.Cadence,
    'hr': types.HeartRate,
    'kph': types.Speed._from_kph,
    'lap': types.LapCounter,
    'lat': types.Latitude,
    'lon': types.Longitude,
    'metres': types.Distance._from_discrete,
    'temp': types.Temperature,
    'watts': types.Power,
}


class SRMHeader:
    __slots__ = ('days_since_1880', 'wheel_circum', 'recording_interval',
                 'block_count', 'marker_count', 'comment_len', '_comment')

    def __init__(self, srmfile):
        raw = srmfile.read(self.size)
        values = list(unpack(self.fmt, raw))
        values[2] /= values.pop(3)   # recording_interval (seconds; 1 / Hz)

        for name, value in zip(self.__slots__, values):
            setattr(self, name, value)

    @property
    def fmt(self):
        return '<2H2B2HxB' + '70s'

    @property
    def size(self):
        return calcsize(self.fmt)

    @property
    def comment(self):
        return self._comment.decode().rstrip('\x00')

    @property
    def date(self):
        return DATETIME_1880 + timedelta(days=self.days_since_1880)


class SRMMarker:
    __slots__ = ('_comment', 'active', 'start', 'end', 'average_watts',
                 'average_hr', 'average_cadence', 'average_speed', 'pwc150')

    def __init__(self, srmfile):
        fmt = self.fmt(srmfile.version)
        raw = srmfile.read(calcsize(fmt))
        values = unpack(fmt, raw)

        for name, value in zip(self.__slots__, values):
            setattr(self, name, value)

        # Data fixup: make sure markers are consistently one-indexed.
        self.start = max(self.start, 1)
        self.end = max(self.end, 1)
        # Data fixup: some srmwin versions wrote markers with start > end.
        self.start, self.end = sorted([self.start, self.end])

    @staticmethod
    def fmt(version):
        comment_len = 3 if version < 6 else 255
        fmt = '<%ds' % comment_len
        fmt += 'B7H' if version < 9 else 'B2L5H'
        return fmt

    @property
    def comment(self):
        return self._comment.decode().rstrip('\x00')

    @property
    def indices(self):
        """start and stop attributes, zero-indexed."""
        return {self.start - 1, self.end - 1}   # one-indexed!


class SRMBlock:
    __slots__ = ('sec_since_midnight', 'chunk_count', 'end')

    def __init__(self, srmfile):
        fmt = self.fmt(srmfile.version)
        raw = srmfile.read(calcsize(fmt))
        hsec_since_midnight, self.chunk_count = unpack(fmt, raw)

        # hsec to sec.
        self.sec_since_midnight = timedelta(seconds=hsec_since_midnight / 100)

        self.end = None   # set later

    @staticmethod
    def fmt(version):
        return '<L' + ('H' if version < 9 else 'L')


class SRMChunk:
    __slots__ = ('watts', 'cad', 'hr', 'kph', 'alt', 'temp', 'metres',
                 'lat', 'lon')

    def __init__(self, srmfile):

        self.lat, self.lon = nan, nan   # sensible default

        if srmfile.version < 7:
            self.watts, self.kph = self.compact_power_speed(srmfile)
            self.cad, self.hr = unpack('<BB', srmfile.read(2))
            self.alt, self.temp = nan, nan
        else:
            values = unpack('<HBBllh', srmfile.read(14))
            for name, value in zip(self.__slots__, values):
                setattr(self, name, value)

            if srmfile.version == 9:
                latlon = unpack('<ll', srmfile.read(8))
                self.lat, self.lon = (l * 180 / 0x7fffffff for l in latlon)

            self.temp *= 0.1
            self.kph = 0 if self.kph < 0 else self.kph * 3.6 / 1000

        # Need to *make* a distance field (recording interval is in sec)
        self.metres = srmfile.recording_interval * self.kph / 3.6

    @staticmethod
    def compact_power_speed(srmfile):
        pwr_spd = unpack('<3B', srmfile.read(3))
        watts = (pwr_spd[1] & 0x0f) | (pwr_spd[2] << 0x4)
        kph = ((pwr_spd[1] & 0xf0) << 3 | (pwr_spd[0] & 0x7f)) * 3 / 26
        return watts, kph

    def __iter__(self):
        for name in self.__slots__:
            yield name, getattr(self, name)

    def as_dict(self):
        return {name: value for name, value in self}


@contextmanager
def open_srm(file_path):
    reader = open(file_path, 'rb')

    magic = reader.read(4).decode()
    if magic[:3] != 'SRM':
        raise InvalidFileError("this doesn't look like an srm file!")
    reader.version = int(magic[3])

    yield reader
    reader.close()


@drydoc.gen_records
def gen_records(file_path):
    with open_srm(file_path) as srmfile:
        header = SRMHeader(srmfile)
        srmfile.recording_interval = header.recording_interval  # useful!

        markers = deque(    # deque for popleft()
            SRMMarker(srmfile) for __ in range(header.marker_count + 1))

        blocks = deque(
            SRMBlock(srmfile) for __ in range(header.block_count))

        block_ends = accumulate(block.chunk_count for block in blocks)
        for block, end in zip(blocks, block_ends):
            setattr(block, 'end', end)

        # Calibration data
        zero, slope = unpack('<2H', srmfile.read(4))

        data_count_fmt = '<%sx' % ('H' if srmfile.version < 9 else 'L')
        data_count, = unpack(data_count_fmt,
                             srmfile.read(calcsize(data_count_fmt)))

        # data_count might overflow at 64k, so use sum from blocks instead
        data_count = sum(block.chunk_count for block in blocks)

        # Start generating the chunks (i.e. records)
        # ------------------------------------------

        # Stuff for creating timestamps
        current_block = blocks.popleft()
        timestamp = header.date + current_block.sec_since_midnight
        rec_int = timedelta(seconds=header.recording_interval)

        # Stuff for creating a lap counter
        lap = 0
        new_lap_i = deque(
            reduce(ior, (marker.indices for marker in markers)))
        next_lap = new_lap_i.popleft()

        for i in range(data_count):
            chunk = SRMChunk(srmfile).as_dict()

            if i == current_block.end:
                current_block = blocks.popleft()
                timestamp = header.date + current_block.sec_since_midnight
            else:
                timestamp += rec_int

            if i == next_lap:
                lap += 1
                if new_lap_i:   # are there any more lap triggers?
                    next_lap = new_lap_i.popleft()

            chunk.update(timestamp=timestamp, lap=lap)

            yield chunk


def read_and_format(file_path):
    data = types.ActivityData.from_records(gen_records(file_path))

    timestamps = data.pop('timestamp')
    timeoffsets = timestamps - timestamps[0]

    data._finish_up(column_spec=COLUMN_SPEC,
                    start=timestamps[0], timeoffsets=timeoffsets)

    return data



"""
Decode the Flexible and Interoperable data Transfer (FIT) protocol [1]_.

This subpackage is really a leaner, and somewhat meaner, py3k version of
the python-fitparse package [2]_. The protocol itself is described extensively
in the FIT SDK materials [1]_.

The reading internals---i.e. the protocol implementation---are in the
`_protocol` module. This relies, in turn, on a data module (`_profile`)
wrangled from the "Profile.xlsx" file that comes with the FIT SDK. Do not
try to read this module. It is thoroughly disgusting. The development
version of this package includes the subdirectory "_make_profile" which
better exposes the logic for this module. Read that instead.


.. [1] https://www.thisisant.com/resources/fit
.. [2] https://github.com/dtcooper/python-fitparse/tree/ng

"""
from activityio.fit._reading import read_and_format as read
from activityio.fit._reading import gen_records
